# Miscellaneous cim classes

ACLINE = "cim:ACLineSegment"
BIDDINGAREA = "SN:MarketDeliveryPoint.BiddingArea"
CNODE_CONTAINER = "cim:ConnectivityNode.ConnectivityNodeContainer"
DELIVERYPOINT = "SN:Substation.MarketDeliveryPoint"
EQUIP_CONTAINER = "cim:Equipment.EquipmentContainer"
GEO_REG = "cim:SubGeographicalRegion"
ID_OBJ = "cim:IdentifiedObject"
MARKETCODE = "SN:BiddingArea.marketCode"
OPERATIONAL_LIMIT_SET = "cim:OperationalLimit.OperationalLimitSet"
SUBSTATION = "cim:VoltageLevel.Substation"
SYNC_MACH = "cim:SynchronousMachine"
TC_EQUIPMENT = "cim:Terminal.ConductingEquipment"
TC_NODE = "cim:Terminal.ConnectivityNode"
TR_WINDING = "cim:PowerTransformerEnd"
